/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*                   (c) 1993,1994 by Kaya Memisoglu                       */
/*                                    aka Marc van Shaney                  */
/*                                                                         */
/* Die kommerzielle Nutzung des Source-Codes ohne meine schriftliche       */
/* Genehmigung ist untersagt. Desweiteren hafte ich fr keinerlei          */
/* Schaden den das Programm verursacht.                                    */
/*                                                                         */
/* Geschrieben mit Borland C++ 3.1                                         */
/*     Borland C++ ist eingetragenes Warenzeichen der                      */
/*                                  Borland Inernational INC               */
/*                                                                         */
/*                                                                         */
/* 18.1.1994 - Kaya Memisoglu                                              */
/*                                                                         */
/***************************************************************************/




#include <alloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>
#include <sys\stat.h>

#include "flilib.h"
#include "grafdrv.h"


#define FRAME_SIZE            64032U



extern void *Malloc(size_t);
extern void *FarMalloc(size_t);
extern void Free(void *);




int Open_FLI(char *FName,FLI_Header *H,int Flags)
{
 int f;

 if ((f=open(FName,O_BINARY|O_RDONLY))==NULL)
     return FILE_ERROR;

 lseek(f,0L,SEEK_SET);
 _read(f,H,sizeof(FLI_Header));
 H->Handle=f;
 if (H->Magic!=FLC_MAGIC)
	 H->First_Frame=tell(f);
 else
	 lseek(f,H->First_Frame,SEEK_SET);
 H->Next_Frame=tell(f);
 H->Current_Frame=0;
 H->Flags=Flags;
 H->Cbuf=NULL;

 if ((H->Magic!=FLI_MAGIC) && (H->Magic!=FLC_MAGIC))
      return WRONG_MAGIC;
 if (((unsigned long)H->Width*H->Height)>64500L)
      return WRONG_RESOLUTION;
 if (H->Frames>4000)
      return WRONG_FRAMES;
 return TRUE;
}






int Close_FLI(FLI_Header *H)
{
 if (H->Cbuf!=NULL)
	Free (H->Cbuf);
 close(H->Handle);
 return TRUE;
}






int Read_FLI(FLI_Header *H,Bitmap *B,Palette *Pal)
{
 char *cbuf,*BUF;
 FLI_Frame_Header fh;
 FLI_Chunk_Header *ch;
 int m,ret=TRUE;

 if (H->Current_Frame>=H->Frames)
      {
       H->Current_Frame=0;
       H->Next_Frame=H->First_Frame;
       lseek(H->Handle,H->Next_Frame,SEEK_SET);
      }

 if (!(H->Flags & FLI_EXCLUSIV))
      lseek(H->Handle,H->Next_Frame,SEEK_SET);
 _read(H->Handle,&fh,sizeof(FLI_Frame_Header));

 if (fh.Magic!=FLI_FRAME_MAGIC)
      return WRONG_MAGIC;
 if ((fh.Size-=(long)sizeof(FLI_Frame_Header))==NULL)
     {
      H->Current_Frame++;
      return TRUE;
     }
 if (fh.Size>FLI_CBUF_SIZE)               //This should NEVER happen !
      return MEMORY_ERROR;

 BUF=(char *)Malloc(FLI_CBUF_SIZE);
 _read(H->Handle,BUF,(size_t)fh.Size);
 if (!(H->Flags & FLI_EXCLUSIV))
     H->Next_Frame=tell(H->Handle);

 cbuf=BUF;
 for (m=0;m<fh.Chunks;m++)
      {
       ch=(FLI_Chunk_Header *)cbuf;
       switch (ch->Type)
			{
			 case FLI_DELTA_FLI:
			      UncompDELTA_FLI(cbuf+sizeof(FLI_Chunk_Header),B);
			      break;
			 case FLI_DELTA_FLC:
			      UncompDELTA_FLC(cbuf+sizeof(FLI_Chunk_Header),B);
			      break;
			 case FLI_COLOR:
			      ret=PAL_CHANGED;
			      UncompCLUT(cbuf+sizeof(FLI_Chunk_Header),Pal);
			      break;
			 case FLI_COLOR_24:
			      ret=PAL_CHANGED;
			      UncompCLUT24(cbuf+sizeof(FLI_Chunk_Header),Pal);
			      break;
			 case FLI_BLACK:
			      ClearFLI(B);
			      break;
			 case FLI_BRUN:
			      UncompRLE(cbuf+sizeof(FLI_Chunk_Header),B);
			      break;
			 case FLI_COPY:
			      CopyFLI(cbuf+sizeof(FLI_Chunk_Header),B);
			      break;
			 case FLI_PSTAMP:
			      break;
			 default:
			      Free(BUF);
			      return FILE_CORRUPTED;
			}

       cbuf=(char *)norm_ptr(cbuf+(unsigned int)ch->Size);
      }

 H->Current_Frame++;
 Free(BUF);
 return ret;
}








int Seek_FLI (FLI_Header *H,Bitmap *B,Palette *Pal,int n)
{
 char *cbuf,*BUF;
 FLI_Frame_Header fh;
 FLI_Chunk_Header *ch;
 int m,ret=TRUE;


 if (H->Current_Frame==n)
      return TRUE;
 if ((H->Current_Frame>=H->Frames) || (H->Current_Frame>n))
      {
       H->Current_Frame=0;
       H->Next_Frame=H->First_Frame;
       lseek(H->Handle,H->Next_Frame,SEEK_SET);
      }

 if (n>H->Frames)
      return OUT_OF_RANGE;

 if (!(H->Flags & FLI_EXCLUSIV))
      lseek(H->Handle,H->Next_Frame,SEEK_SET);
 if ((BUF=(char *)Malloc(FLI_CBUF_SIZE))==NULL)
      return MEMORY_ERROR;

 while(H->Current_Frame != n)
      {
       _read(H->Handle,&fh,sizeof(FLI_Frame_Header));
       if (fh.Magic!=FLI_FRAME_MAGIC)
	     return WRONG_MAGIC;
       if ((fh.Size-=(long)sizeof(FLI_Frame_Header))!=NULL)
	     {
	      if (fh.Size>FLI_CBUF_SIZE)
		      return MEMORY_ERROR;
	      if(_read(H->Handle,BUF,(size_t)fh.Size)!=(size_t)fh.Size)
		      {
		       Free(BUF);
		       return DISK_ERROR;
		      }
	      cbuf=BUF;

	      for (m=0;m<fh.Chunks;m++)
		      {
		       ch=(FLI_Chunk_Header *)cbuf;
		       switch (ch->Type)
				{
				 case FLI_DELTA_FLI:
				       UncompDELTA_FLI(cbuf+sizeof(FLI_Chunk_Header),B);
				       break;
				 case FLI_DELTA_FLC:
				       UncompDELTA_FLC(cbuf+sizeof(FLI_Chunk_Header),B);
				       break;
				 case FLI_COLOR:
				       ret=PAL_CHANGED;
				       UncompCLUT(cbuf+sizeof(FLI_Chunk_Header),Pal);
				       break;
				 case FLI_COLOR_24:
				       ret=PAL_CHANGED;
				       UncompCLUT24(cbuf+sizeof(FLI_Chunk_Header),Pal);
				       break;
				 case FLI_BLACK:
				       ClearFLI(B);
				       break;
				 case FLI_BRUN:
				       UncompRLE(cbuf+sizeof(FLI_Chunk_Header),B);
				       break;
				 case FLI_COPY:
				       CopyFLI(cbuf+sizeof(FLI_Chunk_Header),B);
				       break;
				 case FLI_PSTAMP:
				       break;
				}

		       cbuf=(char *)norm_ptr(cbuf+(unsigned int)ch->Size);
		      }
	     }

       H->Current_Frame++;
      }

 H->Next_Frame=tell(H->Handle);
 Free(BUF);
 return ret;
}






int Create_FLI(char *FName,FLI_Header *H,int Flags)
{
 int f;

 memset(H,0,sizeof(FLI_Header));
 if ((f=open(FName,O_RDWR|O_BINARY|O_CREAT|O_TRUNC,S_IREAD|S_IWRITE))==NULL)
       return DISK_ERROR;

 lseek(f,0L,SEEK_SET);
 H->Magic=FLI_MAGIC;
 H->Size=128L;
 H->Next_Frame=128L;
 H->First_Frame=128L;
 H->Width=320;
 H->Height=200;
 H->Bits_a_pixel=8;
 H->Cbuf=NULL;

 _write(f,H,sizeof(FLI_Header));
 H->Handle=f;
 H->Flags=Flags;

 return TRUE;
}






int Write_FLI(
	FLI_Header *fh, 					/* Header inited by fli_create */
        Bitmap *this,Palette *thisPal,
        Bitmap *last,Palette *lastPal)
{
 char *cbuf;
 long fsize;
 char *lpixels;
 Palette *lcmap;
 int err = TRUE;
 int type=FLI_LC;

 fh->Width=this->Width;
 fh->Height=this->Height;
 cbuf = (char *)Malloc(FLI_CBUF_SIZE);
 if (last == NULL)
	{
 	 lpixels = NULL;
	 lcmap = NULL;
	}
    else
	{
	 lpixels = last->Data;
	 lcmap = lastPal;
	}
 if (fh->Frames==0)
	type=FLI_BRUN;


 fsize = FLI_Compress(cbuf, lpixels, lcmap, this->Data, thisPal,type);

 lseek(fh->Handle,fh->Next_Frame,SEEK_SET);
 if (_write(fh->Handle, cbuf,(unsigned int) fsize) != (unsigned int)fsize)
	 err = DISK_ERROR;

 Free(cbuf);
 fh->Size += fsize;
 fh->Frames ++;
 fh->Next_Frame=tell(fh->Handle);
 return(err);
}








int Finish_FLI(FLI_Header *fh, Bitmap *EndMap, Palette *EndPal)
{
 long lastpos;
 Bitmap *firstMap;
 Palette *firstPal;
 int handle=fh->Handle;

 firstPal=(Palette *)Malloc(sizeof(Palette));
 firstMap=(Bitmap *)Malloc(FRAME_SIZE);

 lastpos = fh->Next_Frame;
 lseek (handle, (long)sizeof(FLI_Header), SEEK_SET);
 fh->Next_Frame=(long)sizeof(FLI_Header);
 Read_FLI(fh, firstMap, firstPal);
 lseek (handle, lastpos, SEEK_SET);
 fh->Next_Frame=lastpos;
 Write_FLI(fh, firstMap,firstPal,EndMap,EndPal);
 fh->Frames--;
 lseek (handle, 0L, SEEK_SET);

 if (fh->Cbuf!=NULL)
	Free (fh->Cbuf);
 fh->Flags = (FLI_FINISHED | FLI_LOOPED);
 fh->First_Frame=128;
 fh->Next_Frame=NULL;
 fh->Cbuf=NULL;
 fh->Handle=0;
 fh->Current_Frame=0;

 _write(handle, fh, sizeof(FLI_Header));

 Free(firstMap);
 Free(firstPal);

 close(handle);
 return TRUE;
}









/***************************************************************************/
/*                                                                         */
/* Komprimierungs-Routinen fr das FLI-Format.                             */
/* Der grte Teil stammt von Jim Kent , wurde aber von Kaya Memisoglu     */
/* berarbeitet und angepat                                               */
/*                                                                         */
/*                                                                         */
/***************************************************************************/




static char *fii_brun_comp_line(char *s1,char *cbuf,int count)
{
int wcount;
char *c;
int bcount;
int op_count;
char *start_dif;
int dif_count;

c = cbuf+1;
op_count = 0;
start_dif = s1;
dif_count = 0;
for (;;)
	{
	if (count < 3)
		{
		dif_count += count;
		while (dif_count > 0)
			{
			bcount = (dif_count < FLI_MAX_RUN ? dif_count : FLI_MAX_RUN );
			*c++ = -bcount;
			dif_count -= bcount;
			while (--bcount >= 0)
				*c++ = *start_dif++;
			op_count++;
			}
		*cbuf = op_count;
		return((char *)norm_ptr(c));
		}
	else
		{
		bcount = (count < FLI_MAX_RUN ? count : FLI_MAX_RUN );
		if ((wcount = i86_bsame(s1, bcount)) >= 3)
			{
			while (dif_count > 0)
				{
				bcount = (dif_count < FLI_MAX_RUN ? dif_count : FLI_MAX_RUN );
				*c++ = -bcount;
				dif_count -= bcount;
				while (--bcount >= 0)
					*c++ = *start_dif++;
				op_count++;
				}
			*c++ = wcount;
			*c++ = *s1;
			op_count++;
			s1 += wcount;
			count -= wcount;
			start_dif = s1;
			}
		else
			{
			dif_count++;
			s1++;
			count -= 1;
			}
		}
	}
}







static char *FII_brun(char *s1,char *cbuf,int width,int height)
{
char *c;
char *oc;
long total = 0;

/* store offset of 1st real line and set up for main line-at-a-time loop */
c = (char *)(cbuf);
while (--height >= 0)
	{
	oc = c;
	c = fii_brun_comp_line(s1,c,width);
	total += ptr_to_long(c) - ptr_to_long(oc);
	if (total >= 60000L)
		return(NULL);
	s1 += width;
	}

return(enorm_ptr(c));  //enorm_ptr
}










static char *FII_FullPalette(char *cbuf, Palette *Cmap)
{
 int i;
 *cbuf++ = 1;
 *cbuf++ = 0;
 *cbuf++ = 0;
 *cbuf++ = 0;

 for (i=0;i<256;i++)
    {
     *cbuf++=Cmap->Colour[i].R>>2;
     *cbuf++=Cmap->Colour[i].G>>2;
     *cbuf++=Cmap->Colour[i].B>>2;
    }

 return((char *)norm_ptr(cbuf));
}








long FLI_Compress(
	char *comp_buf, /* Buffer - should be FLI_CBUF_SIZE or bigger */
	char *last_screen, Palette *last_cmap, 	/* Data from previous frame */
	char *this_screen, Palette *this_cmap,	/* Data for this frame */
	int type)				/* FLI_BRUN?  FLI_LCCOMP? */
{
char *c;
FLI_Frame_Header *frame;
FLI_Chunk_Header *chunk;


frame = (FLI_Frame_Header *)comp_buf;
chunk = (FLI_Chunk_Header *)(frame+1);
memset(frame,0,sizeof(FLI_Frame_Header));

/* 1st make the color map chunk */

if (type == FLI_BRUN)
	c = FII_FullPalette((char *)(chunk+1), this_cmap);
    else
	c = FII_CompressPal(last_cmap, this_cmap, (char *)(chunk+1), 256);

chunk->Type = FLI_COLOR;
chunk->Size = ptr_to_long(c) - ptr_to_long(chunk);
if (chunk->Size <= FLI_EMPTY_DCOMP)
	c = (char *)chunk;
    else
	frame->Chunks = 1;

chunk = (FLI_Chunk_Header *)c;

switch (type)
       {
	case FLI_LC:
		c = FII_LineCompare(last_screen, this_screen, (char *)(chunk+1), 320, 200);
		break;
	case FLI_BRUN:
		c = FII_brun(this_screen, (char *)(chunk+1), 320,200);
		break;
       }

if (c == NULL)
       {
	chunk->Size = 64000L+sizeof(FLI_Chunk_Header);
	chunk->Type = FLI_COPY;
	c = norm_ptr((char *)(chunk+1)+64000U);
	memcpy((char *)(chunk+1),this_screen,64000U);
       }
else
       {
	chunk->Type = type;
	chunk->Size = ptr_to_long(c) - ptr_to_long(chunk);
       }

if (chunk->Size <= FLI_EMPTY_DCOMP)
	c = (char *)chunk;
else
	frame->Chunks++;
frame->Magic = FLI_FRAME_MAGIC;
frame->Size = ptr_to_long(c) - ptr_to_long(comp_buf);
return(frame->Size);
}








int Compare_Palette(Palette *p1,Palette *p2,int count)
{
 int i;

 for (i=0;i<count;i++)
	 {
	  if (memcmp(&p1->Colour[i],&p2->Colour[i],3)!=0)
		   return i;
	 }

 return count;
}






int Contrast_Palette(Palette *p1,Palette *p2,int count)
{
 int i;

 for (i=0;i<count;i++)
         {
          if (memcmp(&p1->Colour[i],&p2->Colour[i],3)==0)
                   return i;
         }

 return count;
}






/* FII_CompressPal - compress an rgb triples color map just doing 'skip' compression */
static char *FII_CompressPal(Palette *s1, Palette *s2, unsigned char *cbuf, int count)
{
Palette *Olds2;
USHORT wcount;
char *c;
USHORT op_count;
USHORT dif_count;
USHORT bcount;
USHORT c3;

c = (char *)(cbuf+1);
op_count = 0;
wcount=Compare_Palette(s1,s2,count);
if (wcount == count)
	return(c);	/* stupid way to say got nothing... */
for (;;)
	{
	 c3=Compare_Palette(s1,s2,count);
	 wcount=c3;
	 if ((count -= wcount) == 0)
		goto OUT;	/* same until the end... */
	 *c++ = c3;
	 s1 = (Palette *)&s1->Colour[c3];
	 s2 = (Palette *)&s2->Colour[c3];
	 Olds2=s2;
	 op_count++;

	 /* figure out how long until the next worthwhile "skip" */
	 dif_count = 0;
	 bcount = count;
	 for (;;)
	       {
		wcount = Contrast_Palette(s1,s2,bcount);
		dif_count += wcount;
		s1 = (Palette *)&s1->Colour[wcount];
		s2 = (Palette *)&s2->Colour[wcount];
		bcount -= wcount;
		if (bcount >= 1)
			{
			if ((wcount = Compare_Palette(s1,s2,1)) == 1)
				{
				break;
				}
			else
				{
				 dif_count += 1;
				 s1 = (Palette *)&s1->Colour[1];
				 s2 = (Palette *)&s2->Colour[1];
				 bcount -= 1;
				}
			}
		else
			{
			break;
			}
		}

	*c++ = dif_count;
	count -= dif_count;
	for (;;)
               {
               	if (dif_count == 0)
			break;
		dif_count -= 1;
                *c++=Olds2->Colour[0].R>>2;
                *c++=Olds2->Colour[0].G>>2;
                *c++=Olds2->Colour[0].B>>2;
                Olds2=(Palette *) &Olds2->Colour[1];
		}
	if (count <= 0)
		break;
	}
OUT:
*cbuf = op_count;
return(enorm_ptr(c));
}






#define INERTIA 4

static char *sbrc_line(char *s1,char *s2,char *cbuf,int count)
{
int wcount;
char *c;
int op_count;
// int same_count;
int next_match;
int bcount;

op_count = 0;
c = cbuf+1;
for (;;)
	{
	if (count <= 0)
		{
		goto OUT;
		}
	/* first find out how many bytes to skip... */
	wcount = i86_bcompare(s1, s2, count);
	if ((count -= wcount) <= 0)
		goto OUT;	/* same until the end... */
	/* if skip is longer than 255 have to break it up into smaller ops */
	while (wcount > 255)
		{
		s1 += 255+1;
		s2 += 255;
		wcount -= 255+1;
		/* make dummy copy 1 op */
		*c++ = 255;
		*c++ = 1;
		*c++ = *s2++;
		op_count++;
		}
	/* save initial skip and move screen pointer to 1st different byte */
	*c++ = wcount;
	s1 += wcount;
	s2 += wcount;
	op_count++;

	/* if have skipped to near the end do a literal copy... */
	if (count <= INERTIA)
		{
		*c++ = count;
                memcpy(c,s2,count);
		c += count;
		goto OUT;
		}

	/* now look for a run of same... */
	bcount = count;
	if (bcount > FLI_MAX_RUN)
		bcount = FLI_MAX_RUN;

	wcount = i86_bsame(s2, bcount);
	if (wcount >= INERTIA)	/* it's worth doing a same thing thing */
		{
		next_match = fii_tnskip(s1, s2, wcount,INERTIA);

		if (next_match < wcount) /* if it's in our space and a decent size */
			{			/* we'll cut short same run for the skip */
			wcount = next_match;
			}
		*c++ = -wcount;
		*c++ = *s2;
		s1 += wcount;
		s2 += wcount;
		count -= wcount;
		}
	else	/* doing a literal copy.  What can we do to make it short? */
		{
		/* figure out how long until the next worthwhile "skip" */
		/* Have wcount of stuff we can't skip through. */
		wcount = fii_tnsame(s2,fii_tnskip(s1,s2,bcount,INERTIA-1),INERTIA);
		/* Say copy positive count as lit copy op, and put bytes to copy
		   into the compression buffer */
		*c++ = wcount;
		memcpy(c,s2,wcount);
		s1 += wcount;
		s2 += wcount;
		c += wcount;
		count -= wcount;
		}
	}
OUT:
*cbuf = op_count;
return(norm_ptr(c));
}





static char *FII_LineCompare(char *s1,char *s2,char *cbuf,int width,int height)
{
int skip_count, j, *icbuf; //,lcount
char *c;
char *oc;
unsigned acc;
long total;
unsigned last_real;

/* find out how many lines of s1 and s2 are the same */
acc = (width>>2);	/* SHORTS in line */
j = height;
skip_count = 0;
total = 0;
while (--j >= 0)
	{
	if (i86_dcompare(s1, s2, acc) != acc)
		break;
	s1 += width;
	s2 += width;
	skip_count++;
	}

/* If all same do special case for empty frame*/
if (skip_count == height)
	return((char *)(cbuf+1));

/* store offset of 1st real line and set up for main line-at-a-time loop */
icbuf=(int *)cbuf;
*icbuf++ = skip_count;
height -= skip_count;
c = (char *)(cbuf+4);
last_real = 0;	/* keep track of last moving line */
for (j=1; j<=height;j++)
	{
	oc = c;
	if (i86_dcompare(s1,s2,acc) == acc)	/* whole line is the same */
		{
		 *c++ = 0;	/* set op count to 0 */
		}
	else	/* compress line */
		{
		 c = sbrc_line(s1,s2,c,width);
		 last_real = j;
		}
	total += ptr_to_long(c) - ptr_to_long(oc);
	if (total >= 60000L)
		return(NULL);
	s1 += width;
	s2 += width;
	}
    /* set # of lines in compression to last real, removing empty bottom lines
       from buffer */
*icbuf = last_real;
c -= height-last_real;
return(enorm_ptr(c));
}

