/***************************************************************************/
/*                                                                         */
/*                                                                         */
/*                   (c) 1993,1994 by Kaya Memisoglu                       */
/*                                    aka Marc van Shaney                  */
/*                                                                         */
/* Die kommerzielle Nutzung des Source-Codes ohne meine schriftliche       */
/* Genehmigung ist untersagt. Desweiteren hafte ich fr keinerlei          */
/* Schaden den das Programm verursacht.                                    */
/*                                                                         */
/* Geschrieben mit Borland C++ 3.1                                         */
/*     Borland C++ ist eingetragenes Warenzeichen der                      */
/*                                  Borland Inernational INC               */
/*                                                                         */
/*                                                                         */
/* 18.1.1994 - Kaya Memisoglu                                              */
/*                                                                         */
/***************************************************************************/



#ifndef __FLILIB_H
#define __FLILIB_H

#include <stdio.h>
#include "types.h"
#include "grafdrv.h"
#include "media.h"


#define FLI_CBUF_SIZE   	65500U
#define FLI_MAX_RUN     	127
#define FLI_FINISHED    	1	/* finished writing fli */
#define FLI_LOOPED		2	/* fli has a loop frame */

#define FLI_MAGIC       	0xaf11
#define FLC_MAGIC		0xAF12u	/* Variable rez .FLC type ID */
#define FLI_FRAME_MAGIC     	0xf1fa
#define FLI_PREFIX_MAGIC 	0xF100u
#define FLI_EMPTY_DCOMP 	8L /* Size returned by fii functions
					to indicate no change (8) */

#define FLI_EXCLUSIV        	1

#define FLI_COLOR           	11
#define FLI_DELTA_FLI      	12
#define FLI_LC			12
#define FLI_COLOR_24		4
#define FLI_DELTA_FLC		7
#define FLI_PSTAMP		18
#define FLI_BLACK           	13
#define FLI_BRUN            	15
#define FLI_COPY           	16

#define PAL_CHANGED         	3


typedef struct
	{
	long    Size;
	USHORT  Magic;  					/* = FLIH_MAGIC */
	USHORT  Frames;
	USHORT  Width;
	USHORT  Height;
	USHORT  Bits_a_pixel;
	SHORT   Flags;
	long    Speed;
	int     Handle;		/* Used by player */
	long	created;	/* Date of flic creation. (FLC only.) */
	long	creator;	/* Serial # of flic creator. (FLC only.) */
	long	updated;	/* Date of flic update. (FLC only.) */
	long	updater;	/* Serial # of flic updater. (FLC only.) */
	int   	aspect_dx;	/* Width of square rectangle. (FLC only.) */
	int   	aspect_dy;	/* Height of square rectangle. (FLC only.) */
	char 	reserved2[32];  /* Set to zero. */
	int     Current_Frame;
	char    *Cbuf;          /* compression buffer */
	long 	First_Frame;	/* Offset to frame 1. (FLC only.) */
	long 	Second_Frame;	/* Offset to frame 2. (FLC only.) */
	long 	Next_Frame;
	char 	reserved3[36];	/* Set to zero. */
	} FLI_Header;

typedef struct
	{
	long Size;
	USHORT Magic;		/* = 0xf1fa FLIF_MAGIC */
	SHORT Chunks;
	char pad[8];
	} FLI_Frame_Header;


typedef struct
	{
	long Size;
	SHORT Type;
	} FLI_Chunk_Header;

typedef struct
	{
	long Size;		/* Size of prefix including header. */
	int Magic;	/* Always PREFIX_TYPE. */
	int Chunks;	/* Number of subchunks in prefix. */
	char Reserved[8];/* Always 0. */
	} FLI_Prefix_Header;
#define FLI PREFIX_MAGIC 0xF100u




extern void UncompRLE(char *,Bitmap *);            	/* 1st frame */
extern void UncompDELTA_FLI(char *,Bitmap *n);		/* delta frames */
extern void UncompDELTA_FLC(char *,Bitmap *n);		/* delta frames */
extern void UncompCLUT(char *, Palette *);		/* colors to memory */
extern void UncompCLUT24(char *, Palette *);		/* colors to memory */
extern void CopyFLI(char *,Bitmap *);
extern void ClearFLI(Bitmap *);
extern void *norm_ptr(void *);                   	/* fold offset into address */
extern void *enorm_ptr(void *);
extern void ReduceFIRST(char *,Palette *,int);
extern int ReduceNEXT(char *Old,char *New,Palette *,int);
extern int fii_tnskip(char *s1,char *s2,int bcount,int mustmatch);
extern int fii_tnsame(char *s,int bcount,int mustmatch);

int Read_FLI(FLI_Header *,Bitmap *,Palette *);
int Open_FLI(char *,FLI_Header *,int);
int Close_FLI(FLI_Header *);
int Seek_FLI(FLI_Header *,Bitmap *,Palette *,int n);
int Create_FLI(char *,FLI_Header *,int);
int Write_FLI(FLI_Header *,Bitmap *,Palette *,Bitmap *,Palette *);
int Finish_FLI(FLI_Header *,Bitmap *,Palette *);


int i86_bsame(void *d, int count);	/* how many bytes in a row the same ? */
int i86_wsame(void *d, int count);	/* how many 16-bit-words the same? */
int i86_bcontrast(void *s1, void *s2, int count); /* bytes til 1st match? */
int i86_wcontrast(void *s1,void *s2,int count); /* words til 1st match? */
int i86_bcompare(void *s1,void *s2,int count); /* how many bytes match */
int i86_dcompare(void *s1,void *s2,int count); /* how many words match */


#define ptr_to_long(a)  ( ((long)FP_SEG(a)<<4) + ((long)FP_OFF(a)) )
int Compare_Palette(Palette *,Palette *,int);
int Cotrast_Palette(Palette *,Palette *,int);
char *fii_brun_comp_line(char *,char *,int);
char *FII_Brun(char *,char *,int,int);
char *FII_FullPalette(char *,Palette *);
char *FII_CompressPal(Palette *,Palette *,unsigned char *,int);
char *FII_LineCompare(char *,char *,char *,int ,int);
char *sbrc_line(char *,char *,char *,int);
long FLI_Compress(char *,char *,Palette *,char *,Palette *,int);


#endif