/***************************************************************************
 **
 ** (c) 1993,1994,1995 by Marc van Shaney , aka Kaya Memisoglu
 **
 ** This Source-Code was written by and is copyrighted by Kaya Memisoglu
 ** You may use the whole code or only parts of it in your own programs
 ** whether they are commercial or not. BUT you must give credits for these
 ** parts to me (Kaya Memisoglu  or  Marc van Shaney).
 ** I am not responsible for any damage or loss caused by this program, so
 ** if you success in erasing your harddisk with this code, then it is your
 ** own fault and you should not claim me to be responsible...
 ** I would be very pleased if you also sent me a postcard or even a letter
 ** if you can use this code. My address:
 **
 **        Kaya Memisoglu
 **        Reichenberger Ring 50
 **        63512 Hainburg
 **        Germany
 **
 **
 **
 **/
#include <io.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#include <alloc.h>


#define	TRUE		1
#define FALSE		-1
#define GVE_VERSION	0x0200L


#ifndef __PCX_FORMAT
 #define __PCX_FORMAT

 #define DCX_Magic	       0x3ade68b1L
 #define SCR_Magic	       0x0cd
 #define PCX_Magic             10
 #define PCX_Coding            1
 #define PCX_Version           5
 #define PCX_Pal_Type          0
 #define PCX_Bitplanes         1



 typedef struct {
		char Magic;
		char Version;
		char Pack_Type;
		char Color_Depth;
		int X1,Y1,X2,Y2;
		int X_Resolution;
		int Y_Resolution;
		char EGA_Palette[48];
		char Reserved;
		char Bitplanes;
		int Bytes_per_Line;
		int Palette_Type;
		char Dummy[58];
	       } PCX_Header;
#endif

typedef struct {
		int Shading;
		int Colour;
		int Texture;
		int Highlight;
	       } Material;

typedef struct {
		int  Vertex_Count;
		int  Colour;
		int  Vertex[32];
	       } Vector_Face;

typedef struct {
		long Vertex_Count;
		struct {
			long X;
			long Y;
			long Z;
			int  Index;
		       } Vertex[];
	       } Vertex_List;
typedef struct {
		long Normal_Count;
		struct {
			long X;
			long Y;
			long Z;
			int  Index;
		       } Normal[];
	       } Normal_List;
typedef struct {
		long Textel_Count;
		struct {
			long U;
			long V;
			int  Index;
		       } Texel[];
	       } Texel_List;
typedef struct {
		long Face_Count;
		Vector_Face Face[];
	       } Face_List;
typedef struct {
		char Name[32];
		long Flags;
		long Position[9];
		long Bounding_Box[9];

		int  Vertex_Count;
		int  Faces;
		int  First_Vertex;
		int  Last_Vertex;
		Vertex_List *VList;
		Texel_List *TList;
		Normal_List *NList;
		Face_List *FList;
	       } Vector_Object;

typedef struct {
		int Object_Count;
		Vector_Object *Object[64];
	       } Vector_World;

#ifndef PI
#define PI 3.141592687
#endif

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

typedef signed char  sbyte;
typedef signed short sword;
typedef signed long  sdword;


/****************************************************************************
 **
 ** These are the LZW compression routines...
 ** They are adapted from Steve Rimmers Book "Supercharged Bitmapped
 ** Graphics".
 **
 **/
#define largest_code	4095		/* largest possible code */
#define table_size	5003		/* table dimensions */
#define	min_code_size	8

char masktable[8]={0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
char bittable[8]= {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};

char code_buffer[259];			/* where the codes go */

int oldcode[table_size];		/* the table */
int currentcode[table_size];
char newcode[table_size];

int code_size;
int clear_code;
int eof_code;
int bit_offset;
int byte_offset;
int bits_left;
int max_code;
int free_code;
unsigned int prefix_code;
unsigned int suffix_char;
unsigned long LZW_CSize;		// Compressed Size
unsigned long LZW_USize;		// Uncompressed Size
unsigned long LZW_Position;


/* initialize the code table */
void init_table(void)
{
	int i;

	code_size=min_code_size+1;
	clear_code=(1<<min_code_size);
	eof_code=clear_code+1;
	free_code=clear_code+2;
	max_code=(1<<code_size);

	for(i=0;i<table_size;i++) currentcode[i]=0;
}

/* flush the code buffer */
void flush(FILE *dst,int n)
{
	if (n)	fwrite(code_buffer,n,1,dst);
	LZW_CSize+=(unsigned long)n;
}



/* write a code to the code buffer */
void write_code(FILE *dst,int code)
{
	long temp;

	byte_offset = bit_offset >> 3;
	bits_left = bit_offset & 7;

	if(byte_offset >= 254) {
		flush(dst,byte_offset);
		code_buffer[0] = code_buffer[byte_offset];
		bit_offset = bits_left;
		byte_offset = 0;
	}

	if(bits_left > 0) {
		temp = ((long) code << bits_left) | code_buffer[byte_offset];
		code_buffer[byte_offset]=temp;
		code_buffer[byte_offset+1]=(temp >> 8);
		code_buffer[byte_offset+2]=(temp >> 16);
	}
	else {
		code_buffer[byte_offset] = code;
		code_buffer[byte_offset+1]=(code >> 8);
	}
	bit_offset += code_size;
}




/* write some data to file */
long lwrite(void *s,int size,FILE *dst)
       {
	unsigned char *src;
	int hx,d;
	src=s;
	LZW_USize+=(unsigned long)size;

	if (suffix_char==0x0ffff)
	       {
		suffix_char=(int)*src++;
		size--;

		/* initialize the prefix */
		prefix_code = suffix_char;
	       }


	/* get a character to compress */
	while (size) {

		suffix_char=(int)*src++;
		size--;

		/* derive an index into the code table */
		hx=(prefix_code ^ (suffix_char << 5)) % table_size;
		d=1;

		for(;;) {

			/* see if the code is in the table */
			if(currentcode[hx] == 0) {

				/* if not, put it there */
				write_code(dst,prefix_code);
				d = free_code;

				/* find the next free code */
				if(free_code <= largest_code) {
					oldcode[hx] = prefix_code;
					newcode[hx] = suffix_char;
					currentcode[hx] = free_code;
					free_code++;
				}

				/* expand the code size or scrap the table */
				if(d == max_code) {
					if(code_size < 12) {
						code_size++;
						max_code <<= 1;
					}
					else {
						write_code(dst,clear_code);
						init_table();
					}
				}
				prefix_code = suffix_char;
				break;
			}
			if(oldcode[hx] == prefix_code &&
			   newcode[hx] == suffix_char) {
				prefix_code = currentcode[hx];
				break;
			}
			hx += d;
			d += 2;
			if(hx >= table_size) hx -= table_size;
		}
	}

	return LZW_CSize;
}




long Open_LZW(FILE *dst)
{
	LZW_Position=ftell(dst);
	LZW_CSize=16L;
	LZW_USize=0L;
	fwrite(&LZW_CSize,16,1,dst);

	/* initialize the encoder */
	bit_offset=0;
	suffix_char=0x0ffff;

	init_table();
	write_code(dst,clear_code);

	return LZW_Position;
}





long Close_LZW(FILE *dst)
{
	long buf;

	/* write the prefix code */
	write_code(dst,prefix_code);

	/* and the end of file code */
	write_code(dst,eof_code);

	/* flush the buffer */
	if(bit_offset > 0) flush(dst,(bit_offset+7)/8);

	/* write a zero length block */
	flush(dst,0);

	buf=ftell(dst);
	fseek(dst,LZW_Position,SEEK_SET);
	fwrite(&LZW_CSize,4,1,dst);
	fwrite(&LZW_USize,4,1,dst);
	fseek(dst,buf,SEEK_SET);
	return (LZW_CSize+16L);
}




/***************************************************************************
 **
 ** These are some general file routines
 **
 **/


#define TOK_OBJECTS		1
#define TOK_PALETTE		2
#define TOK_COLOUR		3
#define TOK_IMPORT		4
#define TOK_YES			5
#define TOK_NO			6
#define TOK_LIGHT    		7
#define TOK_DIRECTION		8
#define TOK_POSITION		9
#define TOK_FLAGS		10
#define TOK_BACK_CULLING	11
#define TOK_SHADING		12
#define TOK_NONE		TOK_NO
#define TOK_TRUE		TOK_YES
#define TOK_FALSE		TOK_NO
#define TOK_LAMBERT		14
#define TOK_PHONG		15
#define TOK_FAST_PHONG		16
#define TOK_GOURAUD		17
#define TOK_GLASS		18
#define TOK_GLASS_LAMBERT	19
#define TOK_GLASS_GOURAUD	20
#define TOK_GLASS_PHONG		21
#define TOK_TEXTURE		22
#define TOK_TEXTURE_LAMBERT	23
#define TOK_TEXTURE_GOURAUD	24
#define TOK_TEXTURE_PHONG	25
#define TOK_GLASS_TEXTURE	26
#define TOK_EXPAND_PALETTE	27
#define TOK_HIGHLIGHT		28
#define TOK_WIRE		29
#define TOK_SORTING		30
#define TOK_ANTI_ALIASING	31
#define TOK_HORIZONTAL		32
#define TOK_VERTICAL		33
#define TOK_FULL		34
#define TOK_SHADING_LUT		35
#define TOK_GLASS_LUT		36
#define TOK_CHROME		37
#define TOK_CHROME_LAMBERT	38
#define TOK_CHROME_GOURAUD	39
#define TOK_BLURRING		49
#define TOK_VERTEX		50
#define TOK_NORMALS		51
#define TOK_MATERIAL		52
#define TOK_BRIGHTNESS		53
#define TOK_DARKNESS		54
#define TOK_MACRO		55
#define TOK_VIEWPORT		56
#define TOK_CHROME_PHONG	57
#define TOK_GLASS_CHROME	58
#define TOK_END			70
#define TOK_OPEN_BRACKET	90
#define TOK_CLOSE_BRACKET       91
#define TOK_EQUAL		92

#define TOK_ERROR		-1
#define TOK_EOF			-2




char token[128];
char palette[768];
long Shading=8,Highlight=8;


int Get_Token (FILE *src)
{
 int i;
 int tokenptr=0;
 token[0]=0;

 if (feof(src))
	{
	 printf("\n   Error: Unexpected end of file. Process aborted\n");
	 return TOK_EOF;
	}

 for (;;)
	{
	 i=fgetc(src);
	 if (i==EOF)
		break;

	 if (isspace(i))
		{
		 if (token[0]!=0)
			break;
		}
	 else
		{
		 if ((isalpha(i)) || (i=='_'))
			token[tokenptr++]=i;
		 else
			{
			 if (token[0]!=0)
				{
				 ungetc(i,src);
				 break;
				}

			 token[tokenptr++]=i;
			 break;
			}
		}
	}

 token[tokenptr]=0;
 strupr(token);

 if (strcmp(token,"OBJECTS")==0)
	return TOK_OBJECTS;
 if (strcmp(token,"PALETTE")==0)
	return TOK_PALETTE;
 if (strcmp(token,"COLOUR")==0)
	return TOK_COLOUR;
 if (strcmp(token,"IMPORT")==0)
	return TOK_IMPORT;
 if (strcmp(token,"YES")==0)
	return TOK_YES;
 if (strcmp(token,"NO")==0)
	return TOK_NO;
 if (strcmp(token,"NONE")==0)
	return TOK_NONE;
 if (strcmp(token,"FALSE")==0)
	return TOK_FALSE;
 if (strcmp(token,"TRUE")==0)
	return TOK_TRUE;
 if (strcmp(token,"LIGHT")==0)
	return TOK_LIGHT;
 if (strcmp(token,"DIRECTION")==0)
	return TOK_DIRECTION;
 if (strcmp(token,"VIEWPORT")==0)
	return TOK_VIEWPORT;
 if (strcmp(token,"POSITION")==0)
	return TOK_POSITION;
 if (strcmp(token,"FLAGS")==0)
	return TOK_FLAGS;
 if (strcmp(token,"SHADING")==0)
	return TOK_SHADING;
 if (strcmp(token,"PHONG")==0)
	return TOK_PHONG;
 if (strcmp(token,"NORMALS")==0)
	return TOK_NORMALS;
 if (strcmp(token,"LAMBERT")==0)
	return TOK_LAMBERT;
 if (strcmp(token,"VERTEX")==0)
	return TOK_VERTEX;
 if (strcmp(token,"MACRO")==0)
	return TOK_MACRO;
 if (strcmp(token,"FAST_PHONG")==0)
	return TOK_FAST_PHONG;
 if (strcmp(token,"GOURAUD")==0)
	return TOK_GOURAUD;
 if (strcmp(token,"TEXTURE")==0)
	return TOK_TEXTURE;
 if (strcmp(token,"BLURRING")==0)
	return TOK_BLURRING;
 if (strcmp(token,"MATERIAL")==0)
	return TOK_MATERIAL;
 if (strcmp(token,"BRIGHTNESS")==0)
	return TOK_BRIGHTNESS;
 if (strcmp(token,"DARKNESS")==0)
	return TOK_DARKNESS;
 if (strcmp(token,"CHROME")==0)
	return TOK_CHROME;
 if (strcmp(token,"CHROME_LAMBERT")==0)
	return TOK_CHROME_LAMBERT;
 if (strcmp(token,"CHROME_GOURAUD")==0)
	return TOK_CHROME_GOURAUD;
 if (strcmp(token,"CHROME_PHONG")==0)
	return TOK_CHROME_PHONG;
 if (strcmp(token,"GLASS")==0)
	return TOK_GLASS;
 if (strcmp(token,"GLASS_LAMBERT")==0)
	return TOK_GLASS_LAMBERT;
 if (strcmp(token,"GLASS_GOURAUD")==0)
	return TOK_GLASS_GOURAUD;
 if (strcmp(token,"GLASS_PHONG")==0)
	return TOK_GLASS_PHONG;
 if (strcmp(token,"GLASS_CHROME")==0)
	return TOK_GLASS_CHROME;
 if (strcmp(token,"EXPAND_PALETTE")==0)
	return TOK_EXPAND_PALETTE;
 if (strcmp(token,"GLASS_TEXTURE")==0)
	return TOK_GLASS_TEXTURE;
 if (strcmp(token,"TEXTURE_LAMBERT")==0)
	return TOK_TEXTURE_LAMBERT;
 if (strcmp(token,"TEXTURE_GOURAUD")==0)
	return TOK_TEXTURE_GOURAUD;
 if (strcmp(token,"TEXTURE_PHONG")==0)
	return TOK_TEXTURE_PHONG;
 if (strcmp(token,"END")==0)
	return TOK_END;
 if (strcmp(token,"BACK_CULLING")==0)
	return TOK_BACK_CULLING;
 if (strcmp(token,"ANTI_ALIASING")==0)
	return TOK_ANTI_ALIASING;
 if (strcmp(token,"HORIZONTAL")==0)
	return TOK_HORIZONTAL;
 if (strcmp(token,"VERTICAL")==0)
	return TOK_VERTICAL;
 if (strcmp(token,"FULL")==0)
	return TOK_FULL;
 if (strcmp(token,"SHADING_LUT")==0)
	return TOK_SHADING_LUT;
 if (strcmp(token,"GLASS_LUT")==0)
	return TOK_GLASS_LUT;
 if (strcmp(token,"SORTING")==0)
	return TOK_SORTING;
 if (strcmp(token,"WIRE")==0)
	return TOK_WIRE;
 if (strcmp(token,"HIGHLIGHT")==0)
	return TOK_HIGHLIGHT;
 if (strcmp(token,"{")==0)
	return TOK_OPEN_BRACKET;
 if (strcmp(token,"}")==0)
	return TOK_CLOSE_BRACKET;
 if (strcmp(token,"=")==0)
	return TOK_EQUAL;

 printf("\n   Error: Unknown keyword %s. Process aborted \n",token);

 return TOK_ERROR;
}




int Get_Line (FILE *src,char *dst)
{
 int i;

 if (feof(src))
	 return TOK_EOF;

 for (;;)
	{
	 i=fgetc(src);
	 if (isprint(i))
		*dst++=i;
	 if ((i==13) || (i<0))
		break;
	}

 *dst=0;
 return TRUE;
}




int Get_Equal (FILE *src)
{
 int i;

 i=Get_Token(src);
 if (i==TOK_EQUAL)
	return TRUE;
 else
	{
	 printf("\n   Error: = expected. Process aborted.\n");
	 return FALSE;
	}
}




int Get_Open_Bracket (FILE *src)
{
 int i;

 i=Get_Token(src);
 if (i==TOK_OPEN_BRACKET)
	return TRUE;
 else
	{
	 printf("\n   Error: { expected. Process aborted.\n");
	 return FALSE;
	}
}




long Get_Number(char *src)
{
 long minus=1;
 long out=0;

 while (*src==32) src++;

 if (*src=='-')
	{
	 minus=-1;
	 *src=32;
	 src++;
	}
 while ((*src>='0') && (*src<='9'))
	{
	 out*=10;
	 out+=(long)((*src)-'0');
	 *src=32;
	 src++;
	}

 return out*minus;
}




int Get_Shading(FILE *src)
{
 int tok;

 if ((tok=Get_Token(src))<0)
	return FALSE;
 switch (tok)
	{
	 case TOK_NONE:
		return 8;
	 case TOK_WIRE:
		return 4;
	 case TOK_VERTEX:
		return 96;
	 case TOK_LAMBERT:
		return 12;
	 case TOK_GOURAUD:
		return 16;
	 case TOK_PHONG:
		return 20;
	 case TOK_CHROME:
		return 60;
	 case TOK_CHROME_LAMBERT:
		return 64;
	 case TOK_CHROME_GOURAUD:
		return 68;
	 case TOK_CHROME_PHONG:
		return 72;
	 case TOK_GLASS_CHROME:
		return 84;
	 case TOK_FAST_PHONG:
		return 76;
	 case TOK_GLASS:
		return 24;
	 case TOK_GLASS_LAMBERT:
		return 28;
	 case TOK_GLASS_GOURAUD:
		return 32;
	 case TOK_GLASS_PHONG:
		return 36;
	 case TOK_TEXTURE:
		return 40+0x0100;
	 case TOK_GLASS_TEXTURE:
		return 80+0x0100;
	 case TOK_TEXTURE_LAMBERT:
		return 44+0x0100;
	 case TOK_TEXTURE_GOURAUD:
		return 48+0x0100;
	 case TOK_TEXTURE_PHONG:
		return 52+0x0100;
	 default:
		printf("\n   Error: Wrong SHDAING keyword \n");
		return FALSE;
	}
}


/*****************************************************************************
 **
 ** This is a simple routine for writing a chunk-header. It returns
 ** the start address of the chunk-header
 **
 **/
long Chunk_Position[16];
int  Chunk_Level=0;

long Open_Chunk(FILE *dst,char *T,char *c)
{
 Chunk_Position[Chunk_Level]=ftell(dst);

 fwrite(T,4,1,dst);
 fwrite(c,4,1,dst);
 fwrite(c,4,1,dst);
 Chunk_Level++;
 return (ftell(dst)-12L);
}




long Close_Chunk(FILE *dst)
{
 long buf;
 Chunk_Level--;

 buf=ftell(dst);
 fseek(dst,Chunk_Position[Chunk_Level]+4L,SEEK_SET);
 Chunk_Position[Chunk_Level]=buf-Chunk_Position[Chunk_Level];
 fwrite(&Chunk_Position[Chunk_Level],4,1,dst);
 fseek(dst,buf,SEEK_SET);

 return (ftell(dst));
}




long Write_Main_Header(FILE *dst)
{
 long i=GVE_VERSION;
 Chunk_Level=0;
 fseek(dst,0L,SEEK_SET);
 fwrite("GVEFILE",8,1,dst);
 fwrite(&i,4,1,dst);
 return 12L;
}




/*****************************************************************************
 **
 ** This routine writes exactly one vector object to the file.
 **
 **/
Material Materials[256];
int Write_Object(FILE *dst,Vector_Object *Object)
{
 long col,tex,shad,high,buf,siz;
 char zerobuf[100];
 int n,i,j;
 memset(zerobuf,0,100);
 printf("      Writing object \"%s\" . . .\n",Object->Name);

 Open_Chunk(dst,"OBJS","NONE");
 Open_Chunk(dst,"HEAD","LZW8");
 Open_LZW(dst);
	lwrite(Object,19*4+32,dst);
 Close_LZW(dst);
 Close_Chunk(dst);				// Header
 /***************************************************************************/
 Open_Chunk(dst,"VECT","LZW8");
 Open_LZW(dst);
 lwrite(&Object->VList->Vertex_Count,4,dst);	// Number of vertices
 for (j=0;j<(int)Object->VList->Vertex_Count;j++)
	{
	 lwrite(&Object->VList->Vertex[j],12,dst);
	 if (Object->NList!=NULL)
		lwrite(&Object->NList->Normal[j],12,dst);
	 else
		lwrite(zerobuf,12,dst);
	}
 Close_LZW(dst);
 Close_Chunk(dst);				// Vertices
 /***************************************************************************/
 if (Object->TList!=NULL)
	{
	 Open_Chunk(dst,"TXUV","LZW8");
	 Open_LZW(dst);
	 lwrite(&Object->TList->Texel_Count,4,dst);
	 for (j=0;j<(int)Object->TList->Texel_Count;j++)
		 lwrite(&Object->TList->Texel[j],8,dst);
	 Close_LZW(dst);
	 Close_Chunk(dst);
	}
 /***************************************************************************/
 Open_Chunk(dst,"FACE","LZW8");
 Open_LZW(dst);
 lwrite(&Object->FList->Face_Count,4,dst);	// Number of faces

 for (j=0;j<Object->FList->Face_Count;j++)
	{
	 col=Object->FList->Face[j].Colour;
	 siz=Object->FList->Face[j].Vertex_Count;
	 n=(int)col;
	 if (Materials[n].Shading==0)
		{
		 shad=Shading;
		 tex=col;
		 high=Highlight;
		}
	 else
		{
		 shad=Materials[n].Shading;
		 tex =Materials[n].Texture;
		 high=Materials[n].Highlight;
		 col =Materials[n].Colour;
		}

	 if (shad & 0x0100L)
		 buf=8L*siz+24L;	// Textured object
	 else
		 buf=4L*siz+24L;	// Object size
	 lwrite(&buf,2,dst);
	 lwrite(&shad,1,dst);		// Object type
	 lwrite(&siz,1,dst);		// Number of vertices
	 lwrite(&col,1,dst);		// Colour
	 lwrite(&high,1,dst);		// Phong-exponent
	 lwrite(&tex,2,dst);		// Texture
	 lwrite(zerobuf,16,dst);	// Normal

	 for (i=0;i<siz;i++)
		{
		 lwrite(&Object->FList->Face[j].Vertex[i],2,dst);
		 lwrite(zerobuf,2,dst);
		}
	 if (shad & 0x0100L)
		 lwrite(zerobuf,(unsigned int)(siz*4),dst);
	}

 Close_LZW(dst);
 Close_Chunk(dst);			// Faces
 Open_Chunk(dst,"END","NONE");
 Close_Chunk(dst);
 Close_Chunk(dst);			// Objects
 /***************************************************************************/
 return TRUE;
}


/*****************************************************************************
 **
 ** This routine will test for a face whether it belongs to an object or not
 **
 **/
int Test_Face(Vector_Face *Face,int First,int Last)
{
 int i;

 for (i=0;i<Face->Vertex_Count;i++)
	{
	 if ((Face->Vertex[i]>=First) && (Face->Vertex[i]<=Last))
		return TRUE;
	}
 return FALSE;
}



/*****************************************************************************
 **
 ** These are some routines for 3D-Model object files
 **
 **/
Face_List *Get_Faces_3DO(FILE *src)
{
 char line[200];
 int siz,i,j,entries,col;
 Face_List *List;


 Get_Line(src,line);
 entries=(int)Get_Number(line);
 Get_Line(src,line);

 if((List=malloc(entries*68+4))==NULL)
	return NULL;

 List->Face_Count=entries;

 for (j=0;j<entries;j++)
	{
	 Get_Line(src,line);
	 siz=(int)Get_Number(line);
	 col=(int)Get_Number(line);

	 List->Face[j].Vertex_Count=siz;
	 List->Face[j].Colour=col;

	 for (i=0;i<siz;i++)
		 List->Face[j].Vertex[i]=(int)Get_Number(line)-1;
	}

 return List;
}




Vertex_List *Get_Vertices_3DO(FILE *src)
{
 char line[100];
 Vertex_List *List;
 int j,entries;

 Get_Line(src,line);
 entries=(int)Get_Number(line);
 Get_Line(src,line);

 if((List=malloc(entries*16+4))==NULL)
	return NULL;

 List->Vertex_Count=entries;
 for (j=0;j<entries;j++)
	{
	 Get_Line(src,line);

	 List->Vertex[j].X=Get_Number(line)<<6L;
	 List->Vertex[j].Y=Get_Number(line)<<6L;
	 List->Vertex[j].Z=Get_Number(line)<<6L;
	}

 return List;
}




Vector_World *Get_Objects_3DO(FILE *src)
{
 char line[100];
 int j,entries;
 Vector_World *List;
 Vector_Object *Obj;

 if((List=malloc(sizeof(Vector_World)))==NULL)
	return NULL;

 Get_Line(src,line);
 entries=(int)Get_Number(line);
 Get_Line(src,line);

 memset(List,0,sizeof(Vector_World));
 List->Object_Count=entries;

 for (j=0;j<entries;j++)
	{
	 Get_Line(src,line);
	 if((Obj=malloc(sizeof(Vector_World)))==NULL)
		return NULL;

	 List->Object[j]=Obj;
	 memset(Obj,0,sizeof(Vector_Object));
	 sprintf(Obj->Name,"Object %d",j);
	 Obj->First_Vertex=(int)Get_Number(line)-1;
	 Obj->Last_Vertex=(int)Get_Number(line)-1;
	 Obj->Vertex_Count=Obj->Last_Vertex-Obj->First_Vertex+1;
	}

 return List;
}




int Import_3DO(FILE *dst,char *srcname)
{
 char line[100];
 int i,j,k,face_count;
 FILE *src;
 Vector_World *World=NULL;
 Face_List *FList=NULL;
 Vertex_List *VList=NULL;
 Vector_Object *Obj;

 printf("   Importing 3DO-file %s . . . \n",srcname);

 if ((src=fopen(srcname,"rb"))==NULL)
	{
	 printf("\n   Error opening source file. Process aborted.\n");
	 return FALSE;
	}

 Get_Line (src,line);
 if (strcmp(line,"3D MODEL OBJECTFILE")!=NULL)
	 printf("\n   Error: Source file has wrong magic. Continuing anyway...\n");

 while(Get_Line(src,line)==TRUE)
	{
	 if (strstr(line,"POINTS")!=NULL)
		if((VList=Get_Vertices_3DO(src))==NULL)
			{
			 printf("Error - not enough free memory !\n");
			 fclose(src);
			 return FALSE;
			}
	 if (strstr(line,"FACES")!=NULL)
		if((FList=Get_Faces_3DO(src))==NULL)
			{
			 printf("Error - not enough free memory !\n");
			 fclose(src);
			 return FALSE;
			}
	 if (strstr(line,"OBJECTS")!=NULL)
		if((World=Get_Objects_3DO(src))==NULL)
			{
			 printf("Error - not enough free memory !\n");
			 fclose(src);
			 return FALSE;
			}
	}


 fclose(src);
 if (World==NULL)
	{
	 printf("Error - no objects defined in file !\n");
	 return FALSE;
	}

 for(i=0;i<World->Object_Count;i++)
	{
	 Obj=World->Object[i];

	 Obj->VList=malloc(16*Obj->Vertex_Count+16);
	 Obj->VList->Vertex_Count=Obj->Vertex_Count;
	 for (j=0;j<Obj->Vertex_Count;j++)
		{
		 memcpy(&Obj->VList->Vertex[j],&VList->Vertex[j+Obj->First_Vertex],14);
		 VList->Vertex[j+Obj->First_Vertex].Index=j;
		}

	 face_count=0;
	 for (j=0;j<FList->Face_Count;j++)
		{
		 if (Test_Face(&FList->Face[j],Obj->First_Vertex,Obj->Last_Vertex)==TRUE)
			face_count++;
		}

	 Obj->FList=malloc(face_count*68+4);
	 Obj->FList->Face_Count=face_count;
	 face_count=0;
	 for (j=0;j<FList->Face_Count;j++)
		{
		 if (Test_Face(&FList->Face[j],Obj->First_Vertex,Obj->Last_Vertex)==TRUE)
			{
			 Obj->FList->Face[face_count].Vertex_Count=FList->Face[j].Vertex_Count;
			 Obj->FList->Face[face_count].Colour=FList->Face[j].Colour;
			 for (k=0;k<FList->Face[j].Vertex_Count;k++)
				{
				 Obj->FList->Face[face_count].Vertex[k]=
				      VList->Vertex[FList->Face[j].Vertex[k]].Index;
				}
			 face_count++;
			}
		}

	 Obj->TList=NULL;
	 Obj->NList=NULL;

	 Write_Object(dst,Obj);


	 free(Obj->FList);
	 free(Obj->VList);
	}

 free(World);
 free(VList);
 free(FList);
 return TRUE;
}



/*****************************************************************************
 **
 ** These are some routines for 3D-Studio object files
 **
 **/
typedef struct {
    word    id;
    dword   len;
} TChunkHeader, *PChunkHeader;




void ChunkReader_3DS(FILE *f, int ind, long p) {
    TChunkHeader h;
    int n;
    long pc;

    while (ftell(f) < p) {
	pc = ftell(f);
	if (fread(&h, sizeof(h), 1, f) != 1) return;
	if (h.len == 0) return;
	n = FindChunk(h.id);
	if (n < 0) {
		printf("%*sUnknown 3DS-chunk: 0x%04X, offset 0x%lX, size: %d bytes.\n",
		       ind, "", h.id, pc, h.len);
	    fseek(f, pc + h.len, SEEK_SET);
	} else {
	    pc = pc + h.len;
	    if (ChunkNames[n].func != NULL)
		ChunkNames[n].func(f, ind + 2, pc);
	    else
		ChunkReader(f, ind + 2, pc);
	    fseek(f, pc, SEEK_SET);
	}
	if (ferror(f))
	    break;
    }
}



int Import_3DS(FILE *dst,char *srcname)
{
 long p;
 FILE *src;
 Vector_World *World=NULL;

 printf("   Importing 3DS-file %s . . . \n",srcname);

 if ((src=fopen(srcname,"rb"))==NULL)
	{
	 printf("\n   Error opening source file. Process aborted.\n");
	 return FALSE;
	}
 fseek(src, 0L, SEEK_END);
 p = ftell(src);
 fseek(src, 0L, SEEK_SET);
 ChunkReader_3DS(src, 0L, p);
 fclose(src);

 free(World);
 return TRUE;
}



/***************************************************************************
 **
 ** This function will import a MAP-palette
 **
 **/
long Brightness=100;
long Darkness=100;

int Import_MAP(char *iname)
{
 int i,c1,c2,c3;
 FILE *src;

 if (strchr(iname,'.')==NULL)
	strcat(iname,".MAP");

 printf("   Importing colour-MAP file %s . . . \n",iname);
 if ((src=fopen(iname,"rt"))==NULL)
	{
	 printf("\n   Error opening colour-map file. Process skipped.\n");
	 return FALSE;
	}

 for (i=0;i<256;i++)
	{
	 fscanf(src,"%d%d%d",&c1,&c2,&c3);
	 palette[i*3]=c1 >> 2;
	 palette[i*3+1]=c2 >> 2;
	 palette[i*3+2]=c3 >> 2;
	}

 fclose(src);

 return TRUE;
}



int Import_PAL(char *iname)
{
 FILE *src;

 if (strchr(iname,'.')==NULL)
	strcat(iname,".PAL");

 printf("   Importing colour-PAL file %s . . . \n",iname);
 if ((src=fopen(iname,"rb"))==NULL)
	{
	 printf("\n   Error opening colour-pal file. Process skipped.\n");
	 return FALSE;
	}

 fread(&palette,768,1,src);
 fclose(src);

 return TRUE;
}




int Fit_Colour(long cr,long cg,long cb)
{
 long bestdiff=0x07000000L,newdiff,j;
 int bestcol=0;
 int i;

 for (i=0;i<256;i++)
	{
	 newdiff=((long)palette[i*3])-cr;
	 newdiff*=newdiff;
	 j=((long)palette[i*3+1])-cg;
	 newdiff+=j*j;
	 j=((long)palette[i*3+2])-cb;
	 newdiff+=j*j;

	 if (newdiff<bestdiff)
		{
		 bestdiff=newdiff;
		 bestcol=i;
		}
	}

 return bestcol;
}



int Write_Palette(FILE *dst)
{
 printf("   Writing palette . . .\n");
 Open_Chunk(dst,"PAL","LZW8");
 Open_LZW(dst);
 lwrite(palette,768,dst);
 Close_LZW(dst);
 Close_Chunk(dst);
 return TRUE;
}




int Write_SLUT(FILE *dst)
{
 long i,cr,cg,cb;
 int j;
 char LUT[256];

 printf("   Writing shading LUT . . .     ");
 Open_Chunk(dst,"SLUT","LZW8");
 Open_LZW(dst);

 for (i=1;i<=32;i++)
	{
	 for (j=0;j<256;j++)
		{
		 cr=(((long)palette[3*j])*i*Darkness
		      + ((100L-Darkness)*((long)palette[3*j])<< 5L))/3200L;
		 cg=(((long)palette[3*j+1])*i*Darkness
		      + ((100L-Darkness)*((long)palette[3*j+1])<< 5L))/3200L;
		 cb=(((long)palette[3*j+2])*i*Darkness
		      + ((100L-Darkness)*((long)palette[3*j+2])<< 5L))/3200L;

		 LUT[j]=Fit_Colour(cr,cg,cb);
		}
	 lwrite(LUT,256,dst);
	 printf("\b\b\b\b%.3d%%",100*i/64);
	}

 for (i=31;i>=0;i--)
	{
	 for (j=0;j<256;j++)
		{
		 cr=((((long)palette[3*j])*i - ((i-31L) << 6L))*Brightness
		      + ((100L-Brightness)*((long)palette[3*j])<< 5L))/3200L;
		 cg=((((long)palette[3*j+1])*i - ((i-31L) << 6L))*Brightness
		      + ((100L-Brightness)*((long)palette[3*j+1])<< 5L))/3200L;
		 cb=((((long)palette[3*j+2])*i - ((i-31L) << 6L))*Brightness
		      + ((100L-Brightness)*((long)palette[3*j+2])<< 5L))/3200L;

		 LUT[j]=Fit_Colour(cr,cg,cb);
		}
	 lwrite(LUT,256,dst);
	 printf("\b\b\b\b%.3d%%",100*(32-i)/64+50);
	}

 Close_LZW(dst);
 Close_Chunk(dst);
 printf("\b\b\b\bReady\n");
 return TRUE;
}



int Write_TLUT(FILE *dst)
{
 int i,j,cr,cg,cb;
 char LUT[256];

 printf("   Writing transparency LUT . . .     ");
 Open_Chunk(dst,"TLUT","LZW8");
 Open_LZW(dst);

 for (i=0;i<256;i++)
	{
	 for (j=0;j<256;j++)
		{
		 cr=(((int)palette[3*j])+((int)palette[3*i])) >> 1;
		 cg=(((int)palette[3*j+1])+((int)palette[3*i+1])) >> 1;
		 cb=(((int)palette[3*j+2])+((int)palette[3*i+2])) >> 1;

		 LUT[j]=Fit_Colour(cr,cg,cb);
		}
	 lwrite(LUT,256,dst);
	 printf("\b\b\b\b%.3d%%",100*i/256);
	}

 Close_LZW(dst);
 Close_Chunk(dst);
 printf("\b\b\b\bReady\n");
 return TRUE;
}



/***************************************************************************
 **
 ** This function will import a PCX-file
 **
 **/

long Import_PCX(FILE *dst,char *iname)
{
 int wid,hei,c,d;
 long Size;
 PCX_Header PCX;
 FILE *src;

 if (strchr(iname,'.')==NULL)
	strcat(iname,".PCX");

 printf("   Importing PCX-file %s . . . \n",iname);
 if ((src=fopen(iname,"rb"))==NULL)
	{
	 printf("\n   Error opening PCX-file. Process skipped.\n");
	 return FALSE;
	}

 fread(&PCX,sizeof(PCX_Header),1,src);

 if (PCX.Magic!=PCX_Magic)
     {
      printf("\n   Error reading PCX-file. Process skipped.\n");
      fclose(src);
      return FALSE;
     }

 wid=PCX.X2-PCX.X1+1;
 hei=PCX.Y2-PCX.Y1+1;

 Size=(unsigned long)wid*hei;

 Open_Chunk(dst,"BMP8","LZW8");
 fwrite(&wid,2,1,dst);
 fwrite(&hei,2,1,dst);
 Open_LZW(dst);


 while (Size>0)
	{
	 c=fgetc(src);		// read char
	 if (c>192)
		{
		 d=fgetc(src);
		 c&=63;
		 Size-=(long)c;
		 while (c)
			{
			 lwrite(&d,1,dst);
			 c--;
			}
		}
	 else
		{
		 lwrite(&c,1,dst);
		 Size--;
		}
	}


 Close_LZW(dst);
 Close_Chunk(dst);
 fclose(src);
 return Size+16L;
}




/***************************************************************************
 **
 ** Now follows the script-interpreter
 **
 **/
int Handle_Flags(FILE *src,FILE *dst)
{
 int i,tok;
 long flags=0x20L+0x80L;		// Clear frame , transform normals

 if (Get_Open_Bracket(src)<0)
	return FALSE;

 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_BACK_CULLING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((tok=Get_Token(src))<0)
				return FALSE;
			switch (tok)
				{
				 case TOK_YES:
					flags|=1L;
					break;
				 case TOK_NO:
					flags&=0x0ffffffffL-1L;
					break;
				 default:
					printf("\n   Error: YES/NO expected after BACK_CULLING\n");
					return FALSE;
				}
			break;
		 case TOK_NORMALS:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((tok=Get_Token(src))<0)
				return FALSE;
			switch (tok)
				{
				 case TOK_YES:
					flags|=0x80L;
					break;
				 case TOK_NO:
					flags&=0x0ffffffffL-0x80L;
					break;
				 default:
					printf("\n   Error: YES/NO expected after NORMALS\n");
					return FALSE;
				}
			break;
		 case TOK_BLURRING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((tok=Get_Token(src))<0)
				return FALSE;
			switch (tok)
				{
				 case TOK_YES:
					flags|=16L;
					flags&=0x0ffffffffL-32L;
					break;
				 case TOK_NO:
					flags|=32L;
					flags&=0x0ffffffffL-16L;
					break;
				 default:
					printf("\n   Error: YES/NO expected after BLURRING\n");
					return FALSE;
				}
			break;
		 case TOK_SORTING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((tok=Get_Token(src))<0)
				return FALSE;
			switch (tok)
				{
				 case TOK_NO:
					flags|=2L;
					break;
				 case TOK_YES:
					flags&=0x0ffffffffL-2L;
					break;
				 default:
					printf("\n   Error: YES/NO expected after SORTING\n");
					return FALSE;
				}
			break;
		 case TOK_ANTI_ALIASING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((tok=Get_Token(src))<0)
				return FALSE;
			switch (tok)
				{
				 case TOK_NONE:
					flags&=0x0ffffffffL-12L;
					break;
				 case TOK_HORIZONTAL:
					flags&=0x0ffffffffL-12L;
					flags|=4;
					break;
				 case TOK_VERTICAL:
					flags&=0x0ffffffffL-12L;
					flags|=8;
					break;
				 case TOK_FULL:
					flags|=12;
					break;
				 default:
					printf("\n   Error: NONE/HORIZONTAL/VERTICAL/FULL expected after ANTI_ALIASING\n");
					return FALSE;
				}
			break;
		 case TOK_CLOSE_BRACKET:
			Open_Chunk(dst,"FLAG","NONE");
			fwrite(&flags,4,1,dst);
			Close_Chunk(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in FLAGS: %s\n",token);
			return FALSE;
		}
	}

 return FALSE;
}




int Handle_Palette(FILE *src,FILE *dst)
{
 int i;
 char iname[128];

 if (Get_Open_Bracket(src)<0)
	return FALSE;
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_IMPORT:
			if (Get_Line(src,iname)<0)
				return FALSE;
			strupr(iname);
			if (strstr(iname,".MAP")!=NULL)
				if(Import_MAP(iname)<NULL)   return FALSE;
			if (strstr(iname,".PAL")!=NULL)
				if(Import_PAL(iname)<NULL)   return FALSE;
			break;
		 case TOK_BRIGHTNESS:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Brightness=Get_Number(iname);
			break;
		 case TOK_DARKNESS:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Darkness=Get_Number(iname);
			break;
		 case TOK_CLOSE_BRACKET:
			Write_Palette(dst);
			Write_SLUT(dst);
			Write_TLUT(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in PALETTE: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}





int Handle_Texture(FILE *src,FILE *dst)
{
 int i;
 char iname[128];

 if (Get_Open_Bracket(src)<0)
	return FALSE;
 Open_Chunk(dst,"TEXT","NONE");
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_IMPORT:
			if (Get_Line(src,iname)<0)
				return FALSE;
			Import_PCX(dst,iname);
			break;
		 case TOK_CLOSE_BRACKET:
			Open_Chunk(dst,"END","NONE");
			Close_Chunk(dst);
			Close_Chunk(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in TEXTURE: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}




int Handle_Light(FILE *src,FILE *dst)
{
 int i;
 char line[100];
 struct {
	 long DX;
	 long DY;
	 long DZ;
	 long PX;
	 long PY;
	 long PZ;
	 long Reserved[6];
	} LVector;

 memset(&LVector,0,sizeof(LVector));
 if (Get_Open_Bracket(src)<0)
	return FALSE;
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_DIRECTION:
			if (Get_Equal(src)<0)
				return FALSE;
			Get_Line(src,line);
			LVector.DX=Get_Number(line)<<2L;
			LVector.DY=Get_Number(line)<<2L;
			LVector.DZ=Get_Number(line)<<2L;
			break;
		 case TOK_POSITION:
			if (Get_Equal(src)<0)
				return FALSE;
			Get_Line(src,line);
			LVector.PX=Get_Number(line)<<10L;
			LVector.PY=Get_Number(line)<<10L;
			LVector.PZ=Get_Number(line)<<10L;
			break;
		 case TOK_CLOSE_BRACKET:
			Open_Chunk(dst,"LIGH","NONE");
			fwrite(&LVector,48,1,dst);
			Close_Chunk(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in LIGHT: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}






int Handle_View(FILE *src,FILE *dst)
{
 int i;
 char line[100];
 struct {
	 long DX;
	 long DY;
	 long DZ;
	 long PX;
	 long PY;
	 long PZ;
	 long Reserved[6];
	} VVector;

 memset(&VVector,0,sizeof(VVector));
 if (Get_Open_Bracket(src)<0)
	return FALSE;
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_DIRECTION:
			if (Get_Equal(src)<0)
				return FALSE;
			Get_Line(src,line);
			VVector.DX=Get_Number(line)<<2L;
			VVector.DY=Get_Number(line)<<2L;
			VVector.DZ=Get_Number(line)<<2L;
			break;
		 case TOK_POSITION:
			if (Get_Equal(src)<0)
				return FALSE;
			Get_Line(src,line);
			VVector.PX=Get_Number(line)<<10L;
			VVector.PY=Get_Number(line)<<10L;
			VVector.PZ=Get_Number(line)<<10L;
			break;
		 case TOK_CLOSE_BRACKET:
			Open_Chunk(dst,"VIEW","NONE");
			fwrite(&VVector,48,1,dst);
			Close_Chunk(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in VIEWPORT: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}




int Handle_Materials(FILE *src,FILE *dst)
{
 int i,n=0;
 char iname[128];

 if (Get_Open_Bracket(src)<0)
	return FALSE;
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_MACRO:
			if (Get_Line(src,iname)<0)
				return FALSE;
			n=(int)Get_Number(iname);
			break;
		 case TOK_SHADING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((Materials[n].Shading=Get_Shading(src))<0)
				return FALSE;
			break;
		 case TOK_HIGHLIGHT:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Materials[n].Highlight=(int)Get_Number(iname);
			break;
		 case TOK_COLOUR:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Materials[n].Colour=(int)Get_Number(iname);
			break;
		 case TOK_TEXTURE:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Materials[n].Texture=(int)Get_Number(iname);
			break;
		 case TOK_CLOSE_BRACKET:
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in MATERIAL: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}





int Handle_Objects(FILE *src,FILE *dst)
{
 int i;
 char iname[128];

 if (Get_Open_Bracket(src)<0)
	return FALSE;
 while ((i=Get_Token(src))>0)
	{
	 switch (i)
		{
		 case TOK_SHADING:
			if (Get_Equal(src)<0)
				return FALSE;
			if ((Shading=Get_Shading(src))<0)
				return FALSE;
			break;
		 case TOK_IMPORT:
			if (Get_Line(src,iname)<0)
				return FALSE;
			strupr(iname);
			if (strstr(iname,".3DO")!=NULL)
				Import_3DO(dst,iname);
			else if (strstr(iname,".3DS")!=NULL)
				Import_3DS(dst,iname);
			else
				printf("\n   Error: Unsupported object file %s\n",iname);
			break;
		 case TOK_HIGHLIGHT:
			if (Get_Equal(src)<0)
				return FALSE;
			if (Get_Line(src,iname)<0)
				return FALSE;
			Highlight=Get_Number(iname);
			break;
		 case TOK_CLOSE_BRACKET:
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in OBJECTS: %s\n",token);
			return FALSE;
		}
	}
 return FALSE;
}









int Read_Script(FILE *h,FILE *dst)
{
 int i;

 while ((i=Get_Token(h))>=0)
	{
	 switch (i)
		{
		 case TOK_PALETTE:
			if (Handle_Palette(h,dst)<0)
				return FALSE;
			break;
		 case TOK_FLAGS:
			if (Handle_Flags(h,dst)<0)
				return FALSE;
			break;
		 case TOK_TEXTURE:
			if (Handle_Texture(h,dst)<0)
				return FALSE;
			break;
		 case TOK_OBJECTS:
			if (Handle_Objects(h,dst)<0)
				return FALSE;
			break;
		 case TOK_LIGHT:
			if (Handle_Light(h,dst)<0)
				return FALSE;
			break;
		 case TOK_MATERIAL:
			if (Handle_Materials(h,dst)<0)
				return FALSE;
			break;
		 case TOK_VIEWPORT:
			if (Handle_View(h,dst)<0)
				return FALSE;
			break;
		 case TOK_END:
			Open_Chunk(dst,"END","NONE");
			Close_Chunk(dst);
			return TRUE;
		 default:
			printf("\n   Error: Wrong keyword in MAIN: %s\n",token);
			return FALSE;
		}
	}

 return FALSE;
}



/***************************************************************************
 **
 ** These are the main-routines
 **
 **/

void main(void)
{
 char srcname[128],dstname[128];
 FILE *src,*dst;
 int i;

 for (i=0;i<256;i++)
	Materials[i].Shading=0;

 printf("\nMakeGVE Scene compiler                               (C) 1996 by Kaya Memisoglu\n");
 if (_argc<2)
	{
	 printf("  Usage: MAKEGVE [InFile] {OutFile}                  Compiling date:%s\n",__DATE__);
	 printf("         For InFile the extension .SCN is assumed, for outfile .GVE\n\n");
	 return;
	}

 strcpy(srcname,_argv[1]);
 strupr(srcname);
 if (strchr(srcname,'.')==NULL)
	strcat(srcname,".SCN");
 if ((_argc==2) || (_argv[2] [0]=='-'))
	{
	 strcpy(dstname,srcname);
	 strcpy(strchr(dstname,'.'),".GVE");
	}
 else
	{
	 strcpy(dstname,_argv[2]);
	 strupr(dstname);
	 if (strchr(dstname,'.')==NULL)
		strcat(dstname,".GVE");
	}

 printf("   Input file :%s\n",srcname);
 printf("   Output file:%s\n\n",dstname);

 if ((src=fopen(srcname,"rb"))==NULL)
	{
	 printf("\n   Error opening source file. Process aborted.\n");
	 return;
	}
 if ((dst=fopen(dstname,"wb"))==NULL)
	{
	 printf("\n   Error opening destination file. Process aborted.\n");
	 return;
	}

 Write_Main_Header(dst);
 Read_Script(src,dst);

 fclose(dst);
 fclose(src);
 printf("\nReady !\n");
}